/**
 * <p>Title: Haplotype</p>
 * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: CSIE.NTU</p>
 * @author Yao-Ting Huang
 * @version 1.0
 */

import java.util.*;
import java.io.*;


public class Greedy2 {
  Vector reduction;
  int NumberOfMissingData, size, total=0;
  BufferedWriter bw, out;
  String BlockID;

  public Greedy2(int NumberOfMissingData) {
    this.reduction = new Vector();
    this.NumberOfMissingData = NumberOfMissingData;
    this.init();
  }

  private void init(){
   try{
    this.bw = new BufferedWriter(new FileWriter(new File("Reduction2.txt")));
    this.out = new BufferedWriter(new FileWriter(new File("Greedy2_"+this.NumberOfMissingData+".txt")));
   }
   catch(IOException ee){
    ee.printStackTrace();
     }
  }


  public void Reduce(Vector vt, String BlockID) {
    StringBuffer sb = (StringBuffer) vt.elementAt(0);
    this.BlockID = BlockID;

    //number of bottom nodes
    this.size = vt.size()*(vt.size()-1)/2;

    //reset reduction vector;
    reduction.clear();

    //count the edges for each SNP
    //Each SNP information is stored in Bipartie
    //Eacn SNP is stored in Reduction which represetns the bipartite graph
    for (int i = 0; i < sb.length(); i++) {
      Bipartie gt = new Bipartie();
      gt.locus = i;
      gt.edges = new int[this.size];
      int index = 0;
      for (int j = 0; j < vt.size(); j++) {
        StringBuffer sb1 = (StringBuffer) vt.elementAt(j);
        for (int k = j + 1; k < vt.size(); k++) {
          StringBuffer sb2 = (StringBuffer) vt.elementAt(k);
          if ((sb1.charAt(i) != sb2.charAt(i)) && sb1.charAt(i)!='n' && sb2.charAt(i)!='n') {
          //if ((sb1.charAt(i) != sb2.charAt(i))) {
            gt.edges[gt.count] = index;
            gt.count++;
          }//end of if
          index++;
        }//end of inner for loop
      }
      //store the transformed result
      reduction.addElement(gt);
    }

    //write reduction result
    try{
      for (int l = 0; l < reduction.size(); l++) {
        Bipartie temp = (Bipartie) reduction.elementAt(l);
        bw.write(BlockID+"\t"+temp.count);
        for(int m=0;m<temp.edges.length;m++){
          bw.write("\t"+temp.edges[m]);
        }
        bw.newLine();
        bw.flush();
      }
    }
    catch (Exception ee) {
      ee.printStackTrace();
    }

  }

  public boolean run() {
    int[] table = new int[this.size];
    int tablesize = this.size * (this.NumberOfMissingData + 1);
    int NumberOfTagSNP = 0;
    Vector result = new Vector(); ;

    for (int i = 0; i < table.length; i++) {
      //Set the table with depth m+1
      table[i] = this.NumberOfMissingData + 1;
    }

    while (tablesize > 0) {
      Bipartie bp = findMax(table);
      if (bp == null) {
        //no SNP can cover this column
        //find the column that can not be convered
        for (int o = 0; o < this.size; o++) {
          //this column is not finished
          if (table[o] > 0)
            System.out.println(o);
        }

        return false;
      }
      result.addElement(bp);
      int count = bp.count;
      NumberOfTagSNP++;

      while (count > 0) {
        for (int j = 0; j < bp.count; j++) {
          //this column is not finished
          if (table[bp.edges[j]] > 0) {
            table[bp.edges[j]]--;
            tablesize--;
            count--;
            //this column is finished
          }
          else {
            count--; //this edge is an useless edge
          }
        } //end of for loop
      } //end of inner while loop
    } //end of outer while loop

    this.total = this.total + NumberOfTagSNP;
    //Write result
    try {
      if (NumberOfTagSNP > 0) {
        this.out.write(this.BlockID + "\t" + NumberOfTagSNP);
        for (int k = 0; k < NumberOfTagSNP; k++) {
          Bipartie temp = (Bipartie) result.elementAt(k);
          this.out.write("\t" + temp.locus);
        }
        this.out.newLine();
        this.out.flush();
      }

    } //end of try
    catch (IOException ee) {
      ee.printStackTrace();
    }

    return true;
  }

  private Bipartie findMax(int[] table){
    int max = 0;
    int index = 0;
    boolean found=false;

    for (int i = 0; i < reduction.size(); i++) {
      Bipartie temp = (Bipartie) reduction.elementAt(i);
      int count = 0;

      for (int j = 0; j < temp.count; j++) {
        //this column is not finished
        if (table[temp.edges[j]] > 0){
          count++;
          found = true;
        }
      } //end of for loop

      if (count > max) {
        max = count;
        index = i;
      } //end of if
   }//end of for loop

   //remove the chosen SNP
   if(found)
     return (Bipartie) reduction.remove(index);
   else
     return null;
  }

  public void close(){
   try{
     this.out.write("Total SNPs: "+this.total);
     this.out.flush();
     this.bw.close();
     this.out.close();
   }
   catch(IOException ee){
    ee.printStackTrace();
     }

  }

  /**
   * <p>Title: Haplotype</p>
   *
   * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
   *
   * <p>Copyright: Copyright (c) 2004</p>
   *
   * <p>Company: CSIE.NTU</p>
   * @author Yao-Ting Huang
   * @version 1.0
   */
  private class Bipartie implements Serializable{
    /**
     * Bipartie
     *
     * @param length int
     */

    int count = 0; //number of edges
    int locus;  //locus of the SNP
    int[] edges;
  }
}
